
#include "board.h"
#include "move.h"
#include "checker.h"
#include "enums.h"

#include <fstream>
using namespace std;

void Move::Init(int pfromX, int pfromY, Direction pdir, Checker* pchecker, Board* pon) {
	fromX = pfromX;
	fromY = pfromY;
	dir = pdir;
	checker = pchecker;
	toX = NewX(fromX);
	toY = NewY(fromY);
	on = pon;
}

long Move::Rate(MoveCode code, int mult, bool blockedOnly) {

	//Good...
	const int W_MOVEOK = 1;
	const int W_MULTIJUMP = 1;
	const int W_JUMPABLE = 0;	//not activated [ever]
	const int W_JUMPDONE = 6;

	//Bad...
	const int W_BADMOVEWALL = -5;
	const int W_BADMOVENME = -5;
	const int W_BADSELECT = 0;	//not activated here

	if (blockedOnly && (code != MC_BADMOVE_WALL && code != MC_BADMOVE_NME))
		return 0;

	//Do the rating for the move we just did...
	switch (code) {
		case MC_MOVEOK:
			return mult * W_MOVEOK;
		case MC_MULTIJUMP:
			return mult * W_MULTIJUMP;
		case MC_JUMPABLE:
			return mult * W_JUMPABLE;
		case MC_JUMPDONE:
			return mult * W_JUMPDONE;
		case MC_BADMOVE_WALL:
			return mult * W_BADMOVEWALL;
		case MC_BADMOVE_NME:
			return mult * W_BADMOVENME;
		case MC_BADSELECT:
			return mult * W_BADSELECT;
		case MC_JUMPDONEKING:
		case MC_KING:
			return mult * BASE_WIN;
	}

	return 0;
}

MoveCode Move::Do() {

	if (checker == NULL)
		return MC_BADSELECT;

	MoveCode retval = CalcType();
	if (isBad(retval))
		return retval;

	fromX = checker->getX();
	fromY = checker->getY();
	toX = NewX(fromX);
	toY = NewY(fromY);

	int oldToX, oldToY;
	const Player* lastOwner;

	if (retval == MC_JUMPABLE) {	//newx, newy has a jumpable checker, figure out where we're landing
		lastOwner = on->pieces[toX][toY]->getOwner();
		oldToX = toX;
		oldToY = toY;

		toX = NewX(toX);
		toY = NewY(toY);
		retval = CalcType(lastOwner, false);
		//Our landing is some sort of valid, complete jump
		if (retval == MC_JUMPDONE || retval == MC_MULTIJUMP || retval == MC_JUMPDONEKING) {
			on->pieces[fromX][fromY] = NULL;
			on->DeleteChecker(on->pieces[oldToX][oldToY]);
			on->pieces[toX][toY] = checker->Move(toX, toY);
		}
	} else if (!isBad(retval)) {	//Any other valid move (that is not a jump)
//		case MC_JUMPDONE:
//		case MC_MOVEOK:
//		case MC_JUMPDONEKING:
//		case MC_KING:
		on->pieces[toX][toY] = checker->Move(toX, toY);
		on->pieces[fromX][fromY] = NULL;
	}

	return retval;
}

MoveCode Move::CalcType(const Player* onCheckerLast, bool multicheck) {

	if (toX < 0 || toX >= on->BOARD_WIDTH)
		return MC_BADMOVE_WALL;
	if (toY < 0 || toY >= on->BOARD_HEIGHT)
		return MC_BADMOVE_WALL;

	//NOTE: This gives preference to the non-simple win condition
	//of jumping into king over walking to king

	//The square we're at is not empty
	if (on->pieces[toX][toY] != NULL) {

		//Trying to move into your own checker
		//Can't jump your own checker
		if (checker->getOwner() == on->pieces[toX][toY]->getOwner())
			return MC_BADMOVE_WALL;

		//Can't jump this checker, he's on a border
		if (toX == 0 || toX == on->BOARD_WIDTH-1 || toY == 0 || toY == on->BOARD_HEIGHT-1)
			return MC_BADMOVE_WALL;

		//See if there's a clear space after our jump spot
		int landX = NewX(toX);
		int landY = NewY(toY);
		//FIXME We can probably do without this bounds check
		if (landX >= 0 && landX < on->BOARD_WIDTH && landY >= 0 && landY < on->BOARD_HEIGHT) {
			if (on->pieces[landX][landY] == NULL) {
	//			if (landY == 0 || landY == on->BOARD_HEIGHT-1)
	//				return MC_JUMPDONEKING;	//We know we can't jump any farther
				return MC_JUMPABLE;	//This checker is jumpable. We might be able to jump more or we might be done.
			} else {
				//You're not trying to move into your own checker.
				return MC_BADMOVE_NME;
			}
		}

		//Fallback
		return MC_BADMOVE_WALL;

	} else if (multicheck) {
		//Empty and we're looking for a multijump oppotunity
		//(this means there's no checker touching our empty space)
		//This prevents unnecessary recursion

		if (toY == on->MIN_WIN_Y && checker->getOwner()->getType() == MIN_PLAYER)
			return MC_KING;
		if (toY == on->MAX_WIN_Y && checker->getOwner()->getType() == MAX_PLAYER)
			return MC_KING;

		return MC_MOVEOK;

	} else if (onCheckerLast != NULL) {
		//The space we're at is empty and we saw a checker last

		if (toY == on->MIN_WIN_Y && checker->getOwner()->getType() == MIN_PLAYER)
			return MC_JUMPDONEKING;
		if (toY == on->MAX_WIN_Y && checker->getOwner()->getType() == MAX_PLAYER)
			return MC_JUMPDONEKING;

		//Check for doublejumps in both directions
		Move newmove;
		MoveCode mt;
		for (int i = 0;i < 2;++i) {
		newmove.Init(toX, toY, dir, checker, on);
		mt = newmove.CalcType(NULL, true);
			if (mt == MC_JUMPABLE)
				return MC_MULTIJUMP;
			else if (mt == MC_JUMPDONEKING)
				return MC_MULTIJUMP;
			dir = SwapLeftRight(dir);
		}

		return MC_JUMPDONE;
	}

	if (toY == 0 || toY == on->BOARD_HEIGHT-1)
		return MC_KING;

	//The square is empty, we did not see a checker last time.
	return MC_MOVEOK;
}

int Move::NewX(int x) {
	if (dir == DIR_DN_LEFT || dir == DIR_UP_LEFT)
		return x-1;

	//DIR_DN_RIGHT || DIR_DN_LEFT
	return x+1;
}

int Move::NewY(int y) {
	if (dir == DIR_DN_LEFT || dir == DIR_DN_RIGHT)
		return y+1;

	//DIR_UP_LEFT || DIR_UP_RIGHT
	return y-1;
}

bool Move::CheckKing(int y) {
	if (y == 0 || y == on->BOARD_HEIGHT-1)
		return true;
	return false;
}